#include <stdio.h> 
#include <stdint.h> 
#include <string.h> 
#include <stdlib.h> 
#include <sys/types.h> 
#include <errno.h> 
#include "objmap.h"

typedef struct
{
  unsigned internal:1, byte:23, mask:8;
  void* child[2];
} objmap_internal_t;

typedef struct {
  unsigned internal:1, keylen:31;
  void *key, *data;
} objmap_leaf_t;

static inline int objmap_dir
    (uint8_t *key,uint32_t keylen,objmap_internal_t *node)
{
  uint8_t c;

  if (node->byte<keylen) c= key[node->byte];
                     else c= 0;
  return (1+(node->mask|c))>>8;
}

void* objmap_find(objmap_t *map,void *skey,uint32_t keylen)
{
  uint8_t *key; 
  objmap_internal_t *node;
  objmap_leaf_t *leaf;

  node= map->root;
  if (!node) return 0;

  key= skey;

  while(node->internal)
     node= node->child[objmap_dir(key,keylen,node)];

  leaf= (void*) node;
  if (leaf->keylen!=keylen) return 0;
  if (memcmp(skey, leaf->key, keylen)) return 0;
                                  else return leaf->data;
}


static inline uint32_t objmap_make_mask
    (uint32_t A, uint32_t B)
{
  uint8_t diff;
  diff= A ^ B;
  diff|= diff>>1;
  diff|= diff>>2;
  diff|= diff>>4;
  return (diff&~(diff>>1))^255;
}

static inline objmap_leaf_t* objmap_leaf_node
    (uint8_t *key, uint32_t keylen,void *data)
{
  objmap_leaf_t *N;
  N= malloc(sizeof(*N));
  N->internal= 0;
  N->keylen= keylen;
  N->key= key;
  N->data= data;
  return N;
}

static inline objmap_internal_t* objmap_internal_node
    (uint32_t byte, uint32_t mask)
{
  objmap_internal_t *N;
  N= malloc(sizeof(*N));
  N->internal= 1;
  N->byte= byte;
  N->mask= mask;
  return N;
}

void* objmap_insert(objmap_t* map, void* skey, uint32_t keylen,void *data)
{
  uint8_t *key; 
  objmap_internal_t *node;
  objmap_leaf_t *elt;

  key= skey;

  if (!map->root)
  {
     map->root= objmap_leaf_node(key,keylen,data);
     return 0;
  }

  node= map->root;
  while(node->internal)
    node= node->child[objmap_dir(key,keylen,node)];
  elt= (void*) node;

// Here, elt is the closest leaf node. We will put this under a new
// internal node along with the new leaf node.

  uint32_t crit_byte;
  uint8_t  mask;
  uint8_t  *eltkey;
  int dir_elt;
  objmap_internal_t *node_key;
  uint8_t ebyte,kbyte;

  eltkey= elt->key;

  for(crit_byte=0;
      crit_byte<keylen && crit_byte<elt->keylen
                       && eltkey[crit_byte]==key[crit_byte];
      crit_byte++)
          ;

  if (crit_byte==keylen && elt->keylen==keylen) return elt->data;
  if (crit_byte>=elt->keylen) ebyte= 0;
                         else ebyte= eltkey[crit_byte];
  if (crit_byte>=keylen) kbyte= 0;
                    else kbyte= key[crit_byte];

  mask= objmap_make_mask(ebyte,kbyte);

  dir_elt= (1+(mask|ebyte))>>8;

  node_key= objmap_internal_node(crit_byte,mask);
  node_key->child[1-dir_elt]= objmap_leaf_node(key,keylen, data);

  objmap_internal_t *parent;

  parent= 0;
  node= map->root;

  while(node->internal)
  {
     if (node->byte > crit_byte) break;
     if (node->byte==crit_byte && node->mask > mask) break;
     parent= node;
     node= node->child[objmap_dir(key,keylen,node)];
  }

  node_key->child[dir_elt]= node;
  if (parent) { if (parent->child[0]==node) parent->child[0]= node_key;
                                    else    parent->child[1]= node_key; }
  else map->root= node_key;
  return 0;
}

void *objmap_remove(objmap_t* map, void *skey,uint32_t keylen)
{
  uint8_t *key; 
  void *retv;

  if (!map->root) return 0;

  key= (void*) skey;
   
  objmap_internal_t *node,*sibling,*parent,*gparent;
  objmap_leaf_t *leaf;

  node= map->root;
  gparent= 0;
  sibling= 0;
  parent= 0;

  while(node->internal)
  {
    int dir;
    dir= objmap_dir(key, keylen, node);
    gparent= parent;
    parent= node;
    sibling= node->child[1-dir];
    node= node->child[dir];
  }

  leaf= (void*) node;
  if (leaf->keylen!=keylen) return 0;
  if (memcmp(leaf->key, key, keylen)) return 0; 
  retv= leaf->data;
  free(leaf);

  if (!gparent)
  {
     map->root= sibling;
     if (parent) free(parent);
     return retv;
  }

  if (gparent->child[0]==parent)
    gparent->child[0]= sibling;
  else
    gparent->child[1]= sibling;

  free(parent);

  return retv;
}

void *objmap_destroy(objmap_t *map)
{
  objmap_internal_t *node,*parent,*gparent,*sibling;
  objmap_leaf_t *leaf;
  void *retv;

  if (!map->root) return 0;

  parent= 0;
  gparent= 0;
  sibling= 0;
  node= map->root;
  while(node->internal)
  {
    gparent= parent;
    parent= node;
    sibling= node->child[1];
    node= node->child[0];
  }
  leaf= (void*) node;
  retv= leaf->data;
  free(node);
  if (!gparent) 
  {
    map->root= sibling;
    if (parent) free(parent);
    return retv;
  }
  if (gparent->child[0]==parent)
    gparent->child[0]= sibling;
  else
    gparent->child[1]= sibling;

  free(parent);

  return retv;
}

objmap_t  *objmap_new()
{
  objmap_t *s;
  s= malloc(sizeof(*s));
  s->root= 0;
  return s;
}
