#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include "list_dir.h"

struct sdir {
  dir_t sub;
  int status;
  unsigned int hidehid:1, gettypes:1, dostat:1, 
               hidedev:1, hidedir:1, hidefile:1, hidelink:1;
  int dlen;
  char *buffer;
  int blen;

  DIR *dir;
  struct dirent *ent;
  diren_t result;
};

static int stat_type_to_char(mode_t m)
{
  if (S_ISREG(m)) return 'F';
  if (S_ISDIR(m)) return 'D';
  if (S_ISLNK(m)) return 'L';
  return 'P';
}

static int dirent_type_to_char(int t)
{
  if (t==DT_REG) return 'F';
  if (t==DT_DIR) return 'D';
  if (t==DT_LNK) return 'L';
  return 'P';
}

static int stat_entry(struct sdir *DD)
{
  int R;
  struct stat sbuf;
  if (DD->hidelink)
    R= lstat(DD->result.path, &sbuf);
  else
    R= stat(DD->result.path, &sbuf);
  if (R) return 1;
  DD->result.size= sbuf.st_size;
  DD->result.mtime= sbuf.st_size;
  DD->result.type= stat_type_to_char(sbuf.st_mode);
  return 0;
}

dir_t *open_dir(const char *path,char *options)
{
  struct sdir *DD;
  struct sdir;
  int name_max;

  DD= calloc(1,sizeof(*DD));
  DD->sub.aux= DD;
  DD->status= 'I';

  DD->sub.path= strdup(path);
  DD->dlen= strlen(path);

  DD->dostat= strchr(options, 'S') ? 1: 0;
  DD->hidehid= strchr(options, 'H') ? 1:0;
  DD->gettypes= strchr(options, 'T') ? 1 :0;
  DD->hidedev= strchr(options, 'P') ? 1:0;
  DD->hidedir= strchr(options, 'D') ? 1:0;
  DD->hidefile= strchr(options, 'F') ? 1:0;
  DD->hidelink= strchr(options, 'L') ? 1:0;
  if (DD->hidedev || DD->hidedir || DD->hidefile || DD->hidelink)
     DD->gettypes= 1;

  DD->blen= DD->dlen + 256;
  DD->buffer= malloc(DD->blen);

  DD->dir= opendir(path);
  if (DD->dir==NULL)
    DD->sub.error= strdup(strerror(errno));

  name_max = pathconf(path, _PC_NAME_MAX);
  if (name_max == -1)         /* Limit not defined, or error */
      name_max = 255;         /* Take a guess */
  DD->ent= malloc(sizeof(struct dirent) + name_max + 1);

  return &DD->sub;
}

diren_t *read_dir(dir_t *D)
{
  int eno,stat_done,dl; 
  struct sdir *DD;
  struct dirent *result;

  if (!D) return NULL;
  DD= D->aux; 
  if (DD->status!='I') return NULL;
  if (!DD->dir) return NULL;
  
again:
  eno= readdir_r(DD->dir, DD->ent, &result);
  if (eno)
  {
    DD->sub.error= strdup(strerror(eno));
    DD->status= 'E';
    return NULL;
  } 
  if (!result) { DD->status= 'F'; DD->sub.error= NULL; return NULL; }

  if (!strcmp(DD->ent->d_name, ".") ||
      !strcmp(DD->ent->d_name, "..")) goto again;
  if (DD->hidehid && DD->ent->d_name[0]=='.') goto again;

  dl= strlen(DD->ent->d_name);
  if (DD->dlen+1+dl+1>DD->blen)
  {
    DD->blen= DD->dlen+1+dl+1 + 10;
    DD->buffer= realloc(DD->buffer,DD->blen);
  }
  sprintf(DD->buffer,"%s/%s",DD->sub.path, DD->ent->d_name);
  DD->result.path= DD->buffer;
  DD->result.name= DD->ent->d_name;

  DD->result.mtime= 0; DD->result.size= 0;

  stat_done= 0;
  if (DD->gettypes)
  {
    if (DD->ent->d_type==DT_UNKNOWN)
    { 
      stat_done= 1;
      if (stat_entry(DD)) goto again;
    }
    else
    {
      DD->result.type= dirent_type_to_char(DD->ent->d_type);
    }
    if (DD->hidedev && !strchr("FDL",DD->result.type)) goto again;
    if (DD->hidefile && DD->result.type=='F') goto again;
    if (DD->hidedir && DD->result.type=='D') goto again;
    if (DD->hidelink && DD->result.type=='L') goto again;
  }

  if (DD->dostat && !stat_done)
  {
    if (stat_entry(DD)) goto again;
  }

  return &DD->result;
}

void close_dir(dir_t *D)
{
  struct sdir *DD;
  DD= D->aux;

  if (DD->dir) closedir(DD->dir);
  if (DD->sub.error) free(DD->sub.error);
  free(DD->sub.path);
  free(DD->buffer);
  free(DD->ent);
  free(DD);
}
